@@ -36,6 +36,7 @@ typedef struct {
   PyObject *uid;
   PyObject *gid;
 
+  PyObject *hwaddr;
   PyObject *hwhdr;
   PyObject *payload;
 
@@ -56,7 +57,7 @@ static struct nflog_data * _TupleAsNfad(PyObject *tup);
 static PyObject * _PyLong_AsBigEndian(PyObject *pylong, unsigned char width);
 static PyObject * _ull_AsBigEndian(unsigned long long val, unsigned char width);
 
-static void nd_dealloc(register nflogdataobject *nd) {
+static void nd_dealloc(nflogdataobject *nd) {
   Py_XDECREF(nd->indev);
   Py_XDECREF(nd->physindev);
   Py_XDECREF(nd->outdev);
@@ -86,7 +87,7 @@ static PyObject * _if_indextoname(uint32_t idx) {
 }
 
 // look up device name from index with mock support
-static PyObject * _devname(register nflogdataobject *nd, uint32_t idx) {
+static PyObject * _devname(nflogdataobject *nd, uint32_t idx) {
   if (idx) {
     PyObject *devname;
     // if devnames is non-null, check there first
@@ -175,6 +176,18 @@ PyObject * new_nflogdataobject(struct nflog_data *nfad, PyObject *devnames) {
     Py_INCREF(Py_None);
   }
 
+  // hwaddr
+  struct nfulnl_msg_packet_hw *hw = nflog_get_packet_hw(nfad);
+  if (hw) {
+    if (!(nd->hwaddr = Py_BuildValue("y#", hw->hw_addr, ntohs(hw->hw_addrlen)))) {
+      nd_dealloc(nd);
+      return NULL;
+    }
+  } else {
+    nd->hwaddr = Py_None;
+    Py_INCREF(Py_None);
+  }
+
   char *hwhdr = nflog_get_msg_packet_hwhdr(nfad);
   size_t hwhdr_sz = nflog_get_msg_packet_hwhdrlen(nfad);
   if (!(nd->hwhdr = Py_BuildValue("y#", hwhdr, hwhdr_sz))) {
@@ -198,63 +211,68 @@ PyObject * new_nflogdataobject(struct nflog_data *nfad, PyObject *devnames) {
   return (PyObject *)nd;
 }
 
-static PyObject * nd_get_proto(register nflogdataobject *nd, void *) {
+static PyObject * nd_get_proto(nflogdataobject *nd, void *) {
   return Py_BuildValue("H", nd->proto);
 }
 
-static PyObject * nd_get_hwtype(register nflogdataobject *nd, void *) {
+static PyObject * nd_get_hwtype(nflogdataobject *nd, void *) {
   return Py_BuildValue("H", nd->hwtype);
 }
 
-static PyObject * nd_get_nfmark(register nflogdataobject *nd, void *) {
+static PyObject * nd_get_nfmark(nflogdataobject *nd, void *) {
   return Py_BuildValue("k", nd->nfmark);
 }
 
-static PyObject * nd_get_timestamp(register nflogdataobject *nd, void *) {
+static PyObject * nd_get_timestamp(nflogdataobject *nd, void *) {
   return PyFloat_FromDouble(nd->timestamp);
 }
 
-static PyObject * nd_get_indev(register nflogdataobject *nd, void *) {
+static PyObject * nd_get_indev(nflogdataobject *nd, void *) {
   Py_INCREF(nd->indev);
   return nd->indev;
 }
 
-static PyObject * nd_get_physindev(register nflogdataobject *nd, void *) {
+static PyObject * nd_get_physindev(nflogdataobject *nd, void *) {
   Py_INCREF(nd->physindev);
   return nd->physindev;
 }
 
-static PyObject * nd_get_outdev(register nflogdataobject *nd, void *) {
+static PyObject * nd_get_outdev(nflogdataobject *nd, void *) {
   Py_INCREF(nd->outdev);
   return nd->outdev;
 }
 
-static PyObject * nd_get_physoutdev(register nflogdataobject *nd, void *) {
+static PyObject * nd_get_physoutdev(nflogdataobject *nd, void *) {
   Py_INCREF(nd->physoutdev);
   return nd->physoutdev;
 }
 
-static PyObject * nd_get_uid(register nflogdataobject *nd, void *) {
+static PyObject * nd_get_uid(nflogdataobject *nd, void *) {
   Py_INCREF(nd->uid);
   return nd->uid;
 }
 
-static PyObject * nd_get_gid(register nflogdataobject *nd, void *) {
+static PyObject * nd_get_gid(nflogdataobject *nd, void *) {
   Py_INCREF(nd->gid);
   return nd->gid;
 }
 
-static PyObject * nd_get_hwhdr(register nflogdataobject *nd, void *) {
+static PyObject * nd_get_hwaddr(nflogdataobject *nd, void *) {
+  Py_INCREF(nd->hwaddr);
+  return nd->hwaddr;
+}
+
+static PyObject * nd_get_hwhdr(nflogdataobject *nd, void *) {
   Py_INCREF(nd->hwhdr);
   return nd->hwhdr;
 }
 
-static PyObject * nd_get_payload(register nflogdataobject *nd, void *) {
+static PyObject * nd_get_payload(nflogdataobject *nd, void *) {
   Py_INCREF(nd->payload);
   return nd->payload;
 }
 
-static PyObject * nd_get_prefix(register nflogdataobject *nd, void *) {
+static PyObject * nd_get_prefix(nflogdataobject *nd, void *) {
   Py_INCREF(nd->prefix);
   return nd->prefix;
 }
@@ -271,6 +289,7 @@ static PyGetSetDef nd_getset[] = {
   {"physoutdev", (getter)nd_get_physoutdev, NULL, NULL, NULL},
   {"uid",        (getter)nd_get_uid,        NULL, NULL, NULL},
   {"gid",        (getter)nd_get_gid,        NULL, NULL, NULL},
+  {"hwaddr",     (getter)nd_get_hwaddr,     NULL, NULL, NULL},
   {"hwhdr",      (getter)nd_get_hwhdr,      NULL, NULL, NULL},
   {"payload",    (getter)nd_get_payload,    NULL, NULL, NULL},
   {"prefix",     (getter)nd_get_prefix,     NULL, NULL, NULL},
@@ -295,14 +314,8 @@ static int _set_dev(
       break;
     } else if ((rv = PyUnicode_Compare(ifnames[i], dev)) == 0) {
       break;
-    } else if (rv == -1) {
-      PyGILState_STATE gil = PyGILState_Ensure();
-      // needs GIL
-      if (PyErr_Occurred()) {
-        PyGILState_Release(gil);
-        return -1;
-      }
-      PyGILState_Release(gil);
+    } else if (rv == -1 && _GIL_PyErr_Occurred()) {
+      return -1;
     }
     ++i;
   }
@@ -326,19 +339,29 @@ PyDoc_STRVAR(nd__get_raw_doc,
 "--\n\n"
 "INTENDED FOR DEBUGGING/TESTING ONLY!\n\n"
 "get raw data, can be passed to pass to __new__ to recreate this object\n"
+"\nArgs:\n"
+"    useraw (Union[bool, None]): Whether saved raw data should be used.\n"
+"        Defaults to None, which means 'use if available'.\n If False, a raw\n"
+"        data structure will be synthesized regardless of whether the original\n"
+"        is available. If True, then None will be returned if raw data wasn't\n"
+"        saved at instantiation.\n"
+"\nReturns:\n"
+"   Tuple[Dict[int, str], Tuple[Union[bytes, None]]]: a raw data structure\n"
 "\n"
-"  useraw\n"
-"   If `None`, use saved raw data if available.\n"
-"   If `False`, don\'t use saved raw data.\n"
-"   If `True`, return saved raw data or `None` if unavailable."
 );
-static PyObject * nd__get_raw(register nflogdataobject *nd, PyObject *args) {
+static PyObject * nd__get_raw(nflogdataobject *nd, PyObject *args) {
+  PyObject *pyuseraw = Py_None;
+  if (!PyArg_ParseTuple(args, "|O:_get_raw", &pyuseraw)) { return NULL; }
+  return nd__get_raw_impl((PyObject *)nd, pyuseraw);
+}
+
+PyObject * nd__get_raw_impl(PyObject *o, PyObject *pyuseraw) {
+  nflogdataobject *nd = (nflogdataobject *)o;
   // by default, use the raw data if it exists
   int useraw = !!(nd->devnames);
 
-  PyObject *ret, *devnames, *raw, *item = Py_None;
-  if (!PyArg_ParseTuple(args, "|O:__getnewargs__", &item)) { return NULL; }
-  if (_nflogr_tristate(item, &useraw) != 0) { return NULL; }
+  PyObject *ret, *devnames, *raw, *item;
+  if (_nflogr_tristate(pyuseraw, &useraw) != 0) { return NULL; }
 
   if (!(ret = PyTuple_New(2))) { return NULL; }
 
@@ -418,7 +441,26 @@ static PyObject * nd__get_raw(register nflogdataobject *nd, PyObject *args) {
     ) < 0) { goto nd__get_raw_cleanup; }
   }
 
-  // NFULA_HWADDR - not supported
+  if (nd->hwaddr != Py_None) {
+    // NFULA_HWADDR - construct from nd->hwaddr
+    struct nfulnl_msg_packet_hw hw;
+    memset(&hw, 0, sizeof(hw));
+
+    char *hwaddr_data;
+    Py_ssize_t hwaddr_size;
+    if (PyBytes_AsStringAndSize(nd->hwaddr, &hwaddr_data, &hwaddr_size) != 0) {
+      goto nd__get_raw_cleanup;
+    } else if (hwaddr_size > (Py_ssize_t)sizeof(hw.hw_addr)) {
+      PyErr_SetString(PyExc_ValueError, "hwaddr too large");
+      goto nd__get_raw_cleanup;
+    }
+
+    hw.hw_addrlen = hwaddr_size;
+    memcpy(hw.hw_addr, hwaddr_data, hwaddr_size);
+
+    if (!(item = Py_BuildValue("y#", &hw, sizeof(hw)))) { goto nd__get_raw_cleanup; }
+    PyTuple_SET_ITEM(raw, NFULA_HWADDR-1, item);
+  }
 
   {
     // NFULA_PAYLOAD - construct from nd->payload
@@ -516,7 +558,9 @@ static PyObject * _NfadAsTuple(struct nflog_data *nfad) {
     nfa = nfad->nfa[i];
     if (nfa) {
       PyObject *bytes;
-      if (!(bytes = Py_BuildValue("y#", (((char *)(nfa)) + NFA_LENGTH(0)), nfa->nfa_len))) {
+      // the l3 payload has to have padding because of reasons :-/
+      int len = nfa->nfa_len - (i == (NFULA_PAYLOAD-1) ? 0 : NFA_LENGTH(0));
+      if (!(bytes = Py_BuildValue("y#", (((char *)(nfa)) + NFA_LENGTH(0)), len))) {
         return NULL;
       }
       PyTuple_SET_ITEM(tup, i, bytes);
@@ -586,15 +630,7 @@ static PyObject * _PyLong_AsBigEndian(PyObject *pylong, unsigned char width) {
   if (pylong == Py_None) { Py_RETURN_NONE; }
 
   unsigned long long val = PyLong_AsUnsignedLong(pylong);
-  if (val == ((unsigned long long)-1)) {
-    PyGILState_STATE gil = PyGILState_Ensure();
-    // needs GIL
-    if (PyErr_Occurred()) {
-      PyGILState_Release(gil);
-      return NULL;
-    }
-    PyGILState_Release(gil);
-  }
+  if (val == ((unsigned long long)-1) && _GIL_PyErr_Occurred()) { return NULL; }
 
   return _ull_AsBigEndian(val, width);
 }
@@ -627,7 +663,7 @@ static PyObject * _ull_AsBigEndian(unsigned long long val, unsigned char width)
   return bytes;
 }
 
-PyObject * nd__iter__(register nflogdataobject *nd) {
+PyObject * nd__iter__(nflogdataobject *nd) {
   if (PyType_Ready(&NflogDataItertype) != 0) { return NULL; }
 
   nflogdataiter *iter = PyObject_New(nflogdataiter, &NflogDataItertype);
@@ -638,7 +674,7 @@ PyObject * nd__iter__(register nflogdataobject *nd) {
   return (PyObject *)iter;
 }
 
-PyObject * nd__str__(register nflogdataobject *nd) {
+PyObject * nd__str__(nflogdataobject *nd) {
   // basically equivalent to dict(nd)
   PyObject *dict = PyDict_New();
   if (!dict) { return NULL; }
@@ -653,9 +689,9 @@ PyObject * nd__str__(register nflogdataobject *nd) {
   return repr;
 }
 
-PyObject * nd__repr__(register nflogdataobject *nd) {
+PyObject * nd__repr__(nflogdataobject *nd) {
   if (nd->raw == Py_None) { return nd__str__(nd); }
-  return PyUnicode_FromFormat("%s(%R)", _PyType_Name(Py_TYPE(nd)), nd->raw);
+  return PyUnicode_FromFormat("%s(%R, %R)", _PyType_Name(Py_TYPE(nd)), nd->devnames, nd->raw);
 }
 
 PyObject * nd__new__(PyTypeObject *subtype, PyObject *args, PyObject *) {
@@ -721,17 +757,17 @@ PyTypeObject NflogDatatype {
 };
 
 // iterator helper class
-static void ndi_dealloc(register nflogdataiter *ndi) {
+static void ndi_dealloc(nflogdataiter *ndi) {
   Py_DECREF(ndi->nd);
   PyObject_Del(ndi);
 }
 
-PyObject * ndi__iter__(register nflogdataiter *ndi) {
+PyObject * ndi__iter__(nflogdataiter *ndi) {
   Py_INCREF(ndi);
   return (PyObject *)ndi;
 }
 
-PyObject * ndi__next__(register nflogdataiter *ndi) {
+PyObject * ndi__next__(nflogdataiter *ndi) {
   char *name;
   getter *get;
   PyObject *val;
